@@ -1,4 +1,4 @@
-from numba import jit, njit
+from numba import jit, njit 
 import numpy as np
 from numpy import transpose as t
 import scipy.optimize
@@ -9,11 +9,12 @@ from scipy.special import polygamma
 
 from sklearn.svm import NuSVR
 from sklearn.metrics import mean_squared_error as mse
+from sklearn.model_selection import KFold
 
 from joblib import Parallel, delayed
 import itertools
 import time
-
+import math
 
 __all__ = ['BLADE', 'Framework']
 
@@ -26,8 +27,8 @@ __all__ = ['BLADE', 'Framework']
 # ExpQ : Nsample by Ngene by Ncell
 
 
-#@njit(parallel = True, fastmath = True)
-@njit
+@njit(fastmath=True)
+#@njit
 def ExpF(Beta, Ncell):
     #NSample by Ncell (Expectation of F)
     output = np.empty(Beta.shape)
@@ -37,8 +38,8 @@ def ExpF(Beta, Ncell):
     #return Beta/np.tile(np.sum(Beta, axis=1)[:,np.newaxis], [1, Ncell])
 
 
-#@njit(parallel = True, fastmath = True)
-@njit
+@njit(fastmath=True)
+#@njit
 def ExpQ(Nu, Beta, Omega, Ngene, Ncell, Nsample):
     # Ngene by Nsample (Expected value of Y)
 
@@ -51,8 +52,8 @@ def ExpQ(Nu, Beta, Omega, Ngene, Ncell, Nsample):
 
     return out 
 
-#@njit(parallel = True, fastmath = True)
-@njit
+@njit(fastmath=True)
+#@njit
 def VarQ(Nu, Beta, Omega, Ngene, Ncell, Nsample):
     # Ngene by Nsample (Variance value of Y)
     B0 = np.sum(Beta, axis=1) # Nsample
@@ -101,8 +102,57 @@ def VarQ(Nu, Beta, Omega, Ngene, Ncell, Nsample):
     return VarTerm + CovTerm
 
 
-#@njit(parallel = True, fastmath = True)
-@njit
+def VarQ_debug(Nu, Beta, Omega, Ngene, Ncell, Nsample):
+    # Ngene by Nsample (Variance value of Y)
+    B0 = np.sum(Beta, axis=1) # Nsample
+    Btilda = ExpF(Beta, Ncell) # Nsample by Ncell
+
+    VarB = Btilda * (1-Btilda)
+    for c in range(Ncell):
+        VarB[:,c] = VarB[:,c] / (B0+1)
+
+    # Nsample Ncell Ncell
+    CovB = np.empty((Nsample, Ncell, Ncell))
+    for l in range(Ncell):
+        for k in range(Ncell):
+            CovB[:,l,k] = - Btilda[:,l] * Btilda[:,k] / (1+B0)
+
+
+    # Ngene by Nsample by Ncell by Ncell
+    CovX = np.empty((Ngene, Nsample, Ncell, Ncell))
+    for l in range(Ncell):
+        for k in range(Ncell):
+            for i in range(Nsample):
+                CovX[:,i,l,k] = np.exp(Nu[i,:,k] + Nu[i,:,l] + \
+                        0.5*(np.square(Omega[:,k]) + np.square(Omega[:,l])))
+
+   
+    #Ngene by Nsample
+    #VarTerm = np.dot(np.exp(2*Nu+2*np.square(Omega)), t(VarB+np.square(Btilda)))\
+    #    - np.dot(np.exp(2*Nu + np.square(Omega)), t(np.square(Btilda)))
+
+    VarTerm = np.zeros((Ngene, Nsample))
+    for i in range(Nsample):
+        for c in range(Ncell):
+            VarTerm[:,i] = VarTerm[:,i] + \
+                np.exp(2*Nu[i,:,c] + 2*np.square(Omega)[:,c])*(VarB[i,c] + np.square(Btilda[i,c])) \
+                    - np.exp(2*Nu[i,:,c] + np.square(Omega[:,c]))*(np.square(Btilda[i,c]))
+
+    # Ngene by Ncell
+    CovTerm = np.zeros((Ngene, Nsample))
+    for l in range(Ncell):
+        for k in range(Ncell):
+            if l != k:
+                for i in range(Nsample):
+                    CovTerm[:,i] = CovTerm[:,i] + CovX[:,i,l,k] * CovB[i,l,k]
+    
+     
+    return zip(VarTerm, CovTerm)
+
+
+
+@njit(fastmath=True)
+#@njit
 def Estep_PY(Y, SigmaY, Nu, Omega, Beta, Ngene, Ncell, Nsample):
     Var = VarQ(Nu, Beta, Omega, Ngene, Ncell, Nsample)
     Exp = ExpQ(Nu, Beta, Omega, Ngene, Ncell, Nsample)
@@ -169,7 +219,8 @@ def grad_Nu_python(Y, SigmaY, Nu, Nu0, Beta, Sigma, Omega, Ngene, Ncell, Nsample
     return grad_PX + grad_PY
 
            
-@njit
+@njit(fastmath=True)
+#@njit
 def grad_Nu(Y, SigmaY, Nu, Omega, Beta, Mu0, Alpha0, Beta0, Kappa0, Ngene, Ncell, Nsample):
     # return Nsample by Ngene by Ncell
 
@@ -178,7 +229,7 @@ def grad_Nu(Y, SigmaY, Nu, Omega, Beta, Mu0, Alpha0, Beta0, Kappa0, Ngene, Ncell
     NuExp = np.sum(Nu, 0)/Nsample
 
     Diff = np.zeros((Ngene, Ncell))
-    ExpBetaN = Beta0 + Nsample*(Nsample+3)/2*np.square(Omega) + \
+    ExpBetaN = Beta0 + (Nsample-1)/2*np.square(Omega) + \
             Kappa0*Nsample/(2*(Kappa0 + Nsample)) * (np.square(Omega)/Nsample + np.square(NuExp - Mu0))
 
     for i in range(Nsample):
@@ -268,21 +319,21 @@ def grad_Nu(Y, SigmaY, Nu, Omega, Beta, Mu0, Alpha0, Beta0, Kappa0, Ngene, Ncell
 
     return grad_PX + grad_PY
 
-#@njit(parallel = True, fastmath = True)
-@njit
+@njit(fastmath = True)
+#@njit
 def grad_Omega(Y, SigmaY, Nu, Omega, Beta, Mu0, Alpha0, Beta0, Kappa0, Ngene, Ncell, Nsample):
     # Ngene by Ncell
 
     # gradient of PX (first term)
     AlphaN = Alpha0 + Nsample * 0.5
     NuExp = np.sum(Nu, 0)/Nsample
-    ExpBetaN = Beta0 + Nsample*(Nsample+3)/2*np.square(Omega) + \
+    ExpBetaN = Beta0 + (Nsample-1)/2*np.square(Omega) + \
             Kappa0*Nsample/(2*(Kappa0 + Nsample)) * (np.square(Omega)/Nsample + np.square(NuExp - Mu0))
 
     for i in range(Nsample):
         ExpBetaN = ExpBetaN + 0.5*np.square(Nu[i,:,:] - NuExp)
 
-    Nominator = - AlphaN * Nsample*(Nsample+3)*Omega + Kappa0 /(Kappa0 + Nsample) * Omega
+    Nominator = - AlphaN * (Nsample-1)*Omega + Kappa0 /(Kappa0 + Nsample) * Omega
     grad_PX = Nominator / ExpBetaN
 
 
@@ -366,8 +417,8 @@ def grad_Omega(Y, SigmaY, Nu, Omega, Beta, Mu0, Alpha0, Beta0, Kappa0, Ngene, Nc
 #    return  grad_PX 
     return grad_PX + grad_PY - grad_QX
 
-#@njit(parallel = True, fastmath = True)
-@njit
+@njit(fastmath = True)
+#@njit
 def g_Exp_Beta(Nu, Omega, Beta, B0, Ngene, Ncell, Nsample):
     
     ExpX = np.exp(Nu)
@@ -401,8 +452,8 @@ def g_Exp_Beta_python(Nu, Omega, Beta, B0, Ngene, Ncell, Nsample):
     return g_Exp
 
 
-#@njit(parallel = True, fastmath = True)
-@njit
+@njit(fastmath=True)
+#@njit
 def g_Var_Beta(Nu, Omega, Beta, B0, Ngene, Ncell, Nsample):
 
     B0Rep = np.empty(Beta.shape) # Nsample by Ncell
@@ -495,8 +546,8 @@ def g_Var_Beta(Nu, Omega, Beta, B0, Ngene, Ncell, Nsample):
     return g_Var
 
 
-#@njit(parallel = True, fastmath = True)
-@njit
+@njit(fastmath=True)
+#@njit
 def g_PY_Beta(Nu, Beta, Omega, Y, SigmaY, B0, Ngene, Ncell, Nsample):
 
     # Ngene by Nsample
@@ -526,8 +577,8 @@ def g_PY_Beta(Nu, Beta, Omega, Y, SigmaY, B0, Ngene, Ncell, Nsample):
     
     return grad_PY
 
-#@njit(parallel = True, fastmath = True)
-@njit
+@njit(fastmath=True)
+#@njit
 def g_PY_Beta(Nu, Beta, Omega, Y, SigmaY, B0, Ngene, Ncell, Nsample):
 
     # Ngene by Nsample
@@ -557,13 +608,13 @@ def g_PY_Beta(Nu, Beta, Omega, Y, SigmaY, B0, Ngene, Ncell, Nsample):
     
     return grad_PY
 
-#@njit(parallel = True, fastmath = True)
+#@njit
 @njit(fastmath = True)
 def Estep_PX(Mu0, Nu, Omega, Alpha0, Beta0, Kappa0, Ncell, Nsample):
     NuExp = np.sum(Nu, 0)/Nsample # expected Nu, Ngene by Ncell
     AlphaN = Alpha0 + 0.5*Nsample # Posterior Alpha
 
-    ExpBetaN = Beta0 + Nsample*(Nsample+3)/2*np.square(Omega) + \
+    ExpBetaN = Beta0 + (Nsample-1)/2*np.square(Omega) + \
             Kappa0*Nsample/(2*(Kappa0 + Nsample)) * (np.square(Omega)/Nsample + np.square(NuExp - Mu0))
 
     for i in range(Nsample):
@@ -694,7 +745,7 @@ class BLADE:
             self.Omega = np.zeros((self.Ngene, self.Ncell)) + Omega_Init
  
         if Nu_Init is None:
-            self.Nu = np.zeros((sefl.Nsample, self.Ngene, self.Ncell))
+            self.Nu = np.zeros((self.Nsample, self.Ngene, self.Ncell))
         else:
             self.Nu = Nu_Init
        
@@ -806,7 +857,9 @@ class BLADE:
         PF = self.Estep_PF(Beta)
         QX = self.Estep_QX(Omega)
         QF = self.Estep_QF(Beta)
-        #return PY
+
+        #if not math.isfinite(PX+PY+PF-QX-QF):
+        #    import pdb; pdb.set_trace()
         return PX+PY+PF-QX-QF
 
            
@@ -878,27 +931,31 @@ class BLADE:
 
 
 def Optimize(logY, SigmaY, Mu0, Alpha, Alpha0, Beta0, Kappa0, Nu_Init, Omega_Init, Nsample, Ncell, Init_Fraction):
-    Beta_Init = np.random.gamma(shape=1, size=(Nsample, Ncell)) + t(Init_Fraction) * 10
+    Beta_Init = np.random.gamma(shape=1, size=(Nsample, Ncell)) * 0.1 + t(Init_Fraction) * 10
     obs = BLADE(logY, SigmaY, Mu0, Alpha, Alpha0, Beta0, Kappa0,
-            Nu_Init, Omega_Init, Beta_Init, fix_Nu=False, fix_Omega=True)
+            Nu_Init, Omega_Init, Beta_Init, fix_Nu=True, fix_Omega=True)
     obs.Optimize()
+    
+    #obs.Fix_par['Beta'] = True
     obs.Fix_par['Nu'] = False
     obs.Fix_par['Omega'] = False
     obs.Optimize()
 
+    #obs.Fix_par['Beta'] = False
+    #obs.Optimize()
     return obs
 
 def BLADE_job(X, stdX, Y, Alpha, Alpha0, Kappa0, SY,
-                Init_Fraction, Rep, Crit='E_step', fsel=0.5):
+                Init_Fraction, Rep, Crit='E_step'):
     Ngene, Nsample = Y.shape
     Ncell = X.shape[1]
     
     Mu0 = X
     
-    logY = np.log(Y+10e-6)
-    SigmaY = np.tile(np.std(logY,1)[:,np.newaxis], [1,Nsample]) * SY
+    logY = np.log(Y+1)
+    SigmaY = np.tile(np.std(logY,1)[:,np.newaxis], [1,Nsample]) * SY + 0.1
     Omega_Init = stdX
-    Beta0 = Alpha0 * stdX
+    Beta0 = Alpha0 * np.square(stdX)
     
     Nu_Init = np.zeros((Nsample, Ngene, Ncell))
     for i in range(Nsample):
@@ -925,7 +982,7 @@ def NuSVR_job(X, Y, Nus, sample):
 
 def Framework(X, stdX, Y, Ind_Marker=None, Ind_sample=None, 
         Alphas=[1,10], Alpha0s=[0.1,1,5], Kappa0s=[1,0.5,0.1], SYs=[1,0.3,0.5],
-        Nrep=3, Njob=10, Nrepfinal=10, fsel=0):
+        Nrep=3, Njob=10, Nrepfinal=10, fsel=0, ParallSample=False):
 
     Ngene, Nsample = Y.shape
     Ncell = X.shape[1]
@@ -983,7 +1040,7 @@ def Framework(X, stdX, Y, Ind_Marker=None, Ind_sample=None,
     start = time.time()
     outs = Parallel(n_jobs=Njob, verbose=10)(
             delayed(BLADE_job)(X_small[Ind_use,:], stdX_small[Ind_use,:], Y_small[Ind_use,:], 
-                a, a0, k0, sY, Init_Fraction[:,Ind_sample], Rep=rep, fsel=fsel)
+                a, a0, k0, sY, Init_Fraction[:,Ind_sample], Rep=rep)
                 for a, a0, k0, sY, rep in itertools.product(
                     Alphas, Alpha0s, Kappa0s, SYs, range(Nrep)
                     )
@@ -991,8 +1048,8 @@ def Framework(X, stdX, Y, Ind_Marker=None, Ind_sample=None,
 
     outs, setting = zip(*outs)
     cri = [obs.E_step(obs.Nu, obs.Beta, obs.Omega) for obs in outs]
-    best_obs = outs[np.argmax(cri)]
-    best_set = setting[np.argmax(cri)]
+    best_obs = outs[np.nanargmax(cri)]
+    best_set = setting[np.nanargmax(cri)]
 
     names = ['a_' + str(sett['Alpha']) + '_a0_' + str(sett['Alpha0']) +\
             '_k0_' + str(sett['Kappa0']) + '_sY_' + str(sett['SigmaY']) + '_rep_' + str(sett['Rep'])\
@@ -1008,26 +1065,72 @@ def Framework(X, stdX, Y, Ind_Marker=None, Ind_sample=None,
     if Nsample_small < Nsample or Ngene < Nmarker:
         print("Start inferring per-sample gene expression levels using the entire genes and samples")
 
-        logY = np.log(Y+10e-6) 
+        logY = np.log(Y+1)
         Nu_Init = np.zeros((Nsample, Ngene, Ncell))
         for i in range(Nsample):
             Nu_Init[i,:,:] = X
         Omega_Init = np.square(np.random.normal(0, 0.1, size=(Ngene, Ncell))) + 0.01
         SigmaY = np.tile(np.std(logY,1)[:,np.newaxis], [1,Nsample])*best_set['SigmaY']
 
-        Beta0 = best_set['Alpha0'] * np.square(stdX)
+        # split samples
+        Nfold = Njob // Nrepfinal
+        if Nfold > 1:
+            kf = KFold(Nfold, shuffle=True)
+        else:
+            kf = 0
 
-        final_obs = Parallel(n_jobs=Njob, verbose=10)(
-            delayed(BLADE_job)(X, stdX, Y, best_set['Alpha'], best_set['Alpha0'], best_set['Kappa0'], 
-                best_set['SigmaY'], Init_Fraction, Rep=rep, fsel=fsel)
-                for rep in range(Nrepfinal)
-            )
+        if Nfold < 2 or not ParallSample:
+            final_obs = Parallel(n_jobs=Njob, verbose=10)(
+                delayed(BLADE_job)(X, stdX, Y, best_set['Alpha'], best_set['Alpha0'], best_set['Kappa0'], 
+                    best_set['SigmaY'], Init_Fraction, Rep=rep)
+                    for rep in range(Nrepfinal)
+                )
 
-        outs, setting = zip(*final_obs)
-        cri = [obs.E_step(obs.Nu, obs.Beta, obs.Omega) for obs in outs]
+            outs, setting = zip(*final_obs)
+            cri = [obs.E_step(obs.Nu, obs.Beta, obs.Omega) for obs in outs]
 
-        final_obs = outs[np.argmax(cri)]
+            final_obs = outs[np.argmax(cri)]
 
+        else:
+            print("random split of samples into " + str(Nfold) + " groups.")
+            sampleInd = []
+            for train, test in kf.split(range(Nsample)):
+                sampleInd = sampleInd + [test]
+
+            final_obs = Parallel(n_jobs=Njob, verbose=10)(
+                    delayed(BLADE_job)(X, stdX, Y[:, sampleInd[fold]], best_set['Alpha'], best_set['Alpha0'], best_set['Kappa0'],
+                        best_set['SigmaY'], Init_Fraction[:, sampleInd[fold]], Rep=str(rep)+'_'+str(fold))
+                    for rep, fold in itertools.product(range(Nrepfinal), range(Nfold))
+                )
+
+            outs, setting = zip(*final_obs)
+            cri = [obs.E_step(obs.Nu, obs.Beta, obs.Omega) for obs in outs]
+            fold = [int(sett['Rep'].split('_')[1]) for sett in setting]
+
+            Beta = np.zeros((Nsample, Ncell))
+            Nu = np.zeros((Nsample, Ngene, Ncell))
+            Omega = {}
+            best_outs = {}
+            sample_fold = {}
+            for f in range(Nfold):
+                ind = np.where(np.array(fold) == f)[0]
+                outs_sub = [outs[i] for i in ind]
+                cri_sub = [cri[i] for i in ind]
+                best_out = outs_sub[np.argmax(cri_sub)]
+                best_outs['Fold_'+str(f)] = best_out
+                sample_fold['Fold_'+str(f)] = sampleInd[f]
+                
+                Beta[sampleInd[f],:] = best_out.Beta
+                Nu[sampleInd[f],:,:] = best_out.Nu
+                Omega['Fold_'+str(f)] = best_out.Omega
+
+            final_obs = {
+                    'Beta': Beta,
+                    'Nu': Nu,
+                    'Omega': Omega,
+                    'final_outs': best_outs,
+                    'sample_fold': sample_fold
+                    }
 
     else:
         final_obs = best_obs
